from django.db import models
from django.utils.translation import ugettext_lazy as _


class PaymentMethodManager(models.Manager):
    def get_query_set(self):
        qs = super(PaymentMethodManager, self).get_query_set()
        return qs.filter(is_enabled=True)


class PaymentMethod(models.Model):
    TYPE_BUYSAFE = 0
    TYPE_WEBATM = 1
    TYPE_24PAY = 2
    TYPE_PAYCODE = 3
    TYPE_CHOICES = (
        (TYPE_BUYSAFE, _('BuySafe')),
        (TYPE_WEBATM, _('Web ATM')),
        (TYPE_24PAY, _('24Pay')),
        (TYPE_PAYCODE, _('PayCode'))
    )
    name = models.CharField(max_length=100, verbose_name=_('name'))
    payment_type = models.IntegerField(choices=TYPE_CHOICES)
    content = models.TextField(blank=True, verbose_name=_('content'))
    store_id = models.CharField(
        max_length=11, verbose_name=_('SunTech store ID'),
        help_text=_(
            'Should be exactly 10 characters, starting with a capitalized '
            'alphabet.'
        )
    )
    password = models.CharField(
        max_length=20, verbose_name=_('SunTech password'),
        help_text=_(
            'This should usually be the transaction password, which is an '
            'alphanumeric string with length between 8 to 20.'
        )
    )
    is_enabled = models.BooleanField(default=True, verbose_name=_('enabled'))

    enabled = PaymentMethodManager()
    objects = models.Manager()

    class Meta:
        verbose_name = _('Payment method')
        verbose_name_plural = _('Payment methods')
        ordering = ['-is_enabled', 'payment_type', '-id']

    def save(self, *args, **kwargs):
        super(PaymentMethod, self).save(*args, **kwargs)
        if self.is_enabled:
            (PaymentMethod.objects.exclude(id=self.id)
                          .filter(payment_type=self.payment_type,
                                  is_enabled=True)
                          .update(is_enabled=False))

    def __unicode__(self):
        return self.name



import importlib
from django.conf import settings
from django.http import HttpResponse
from django.db.models.loading import get_model
from django.shortcuts import render
from buysafe.models import PaymentMethod
from buysafe.forms import BuySafeReceiveForm, WebATMReceiveForm


PAYMENT_RECEIVE_FORMS = {
    PaymentMethod.TYPE_BUYSAFE: BuySafeReceiveForm,
    PaymentMethod.TYPE_WEBATM: WebATMReceiveForm
}


def strip_non_number(text):
    result = ''
    for char in text:
        if char.isdigit():
            result += char
    return result


def default_order_info_handler(request, context, **kwargs):
    order_id = kwargs['order_id']
    Order = get_model('shop', 'Order')
    order = Order.objects.get(id=int(order_id))
    return {
        'price': order.total,
        'order_info': unicode(order),
        'order_id': order.id,
        'name': order.billing_name(),
        'phone': strip_non_number(order.billing_detail_phone),
        'email': order.billing_detail_email
    }


def make_response_handler(response_type, content=None):
    if content is not None:
        def f(**kwargs):
            return response_type(content)
    else:
        def f(**kwargs):
            return response_type()
    f._is_response_handler = True
    return f


def call_handler(name, default_handler, **kwargs):
    """
    Call the handler with `name` set in settings.py with provided keyword
    arguments, and fallback to `default_handler` if the name is not set. If
    `default_handler` is none and the name is not set, nothing happends and
    None will be returned.
    """
    try:
        handler_name = getattr(settings, name)
    except AttributeError:
        if default_handler is not None:
            result = default_handler(**kwargs)
        else:
            result = None
    else:
        components = handler_name.split('.')
        module_name, function_name = '.'.join(components[:-1]), components[-1]
        module = importlib.import_module(module_name)
        result = getattr(module, function_name)(**kwargs)
    return result


def call_handler_and_render(name, default_handler, **kwargs):
    result = call_handler(name, default_handler, **kwargs)
    if result is None \
            and getattr(default_handler, '_is_response_handler', False):
        return default_handler(**kwargs)
    if isinstance(result, HttpResponse):
        return result
    return render(kwargs['request'], kwargs['template'], kwargs['context'])


def get_payment_form(payment_type, querydict):
    methods = PaymentMethod.enabled.filter(payment_type=payment_type)
    form = None
    form_type = PAYMENT_RECEIVE_FORMS[payment_type]
    for method in methods:
        password = method.password
        form = form_type(password, querydict)
        if form.is_valid():
            break
        else:
            form = None
    return form



from django import forms


class ChecksumError(forms.ValidationError):
    def __init__(self, expected, got, code=None, params=None):
        message = (
            'Form checksum not valid. Expected {expected}, but got {got}'
            .format(expected=expected, got=got)
        )
        super.__init__(message, code, params)



from django.conf.urls import patterns, url


urlpatterns = patterns(
    'buysafe.views',
    url(r'^entry/(?P<order_id>\d+)/$', 'entry', name='buysafe_pay'),
    url(r'^start/$', 'start', name="buysafe_start"),
    (r'^success/(?P<payment_type>[01])/$', 'success'),
    (r'^fail/(?P<payment_type>[01])/$', 'fail'),
    (r'^check/(?P<payment_type>[01])/$', 'check')
)



from django import forms
from django.contrib import admin
from buysafe.models import PaymentMethod


class PaymentMethodForm(forms.ModelForm):
    class Meta:
        model = PaymentMethod
        widgets = {
            'password': forms.PasswordInput(render_value=True)
        }


class PaymentMethodAdmin(admin.ModelAdmin):
    list_display = ('is_enabled', 'name', 'payment_type', 'store_id', 'content')
    list_editable = ('is_enabled',)
    list_display_links = ('name',)
    form = PaymentMethodForm


admin.site.register(PaymentMethod, PaymentMethodAdmin)



import hashlib
from copy import deepcopy
from django import forms
from django.forms import widgets
from django.utils.http import urlunquote
from buysafe.exceptions import ChecksumError
from buysafe.fields import NumberStringField, SunTechChecksumField
from buysafe.widgets import HiddenIntegerInput


###########
# Gadgets #
###########

def _sha1_hex_upper(*args):
    return hashlib.sha1(''.join([str(v) for v in args])).hexdigest().upper()


#################
# Generic forms #
#################

class SunTechForm(forms.Form):
    """
    A generic form for SunTech transactions. This class overrides `add_prefix`
    to perform a lookup to generate correct form field names for SunTech.
    Subclasses should override class attribute `FIELD_NAME_LOOKUP` to set up
    the lookup table.
    """
    FIELD_NAME_LOOKUP = {}

    def __init__(self, store_password, *args, **kwargs):
        super(SunTechForm, self).__init__(*args, **kwargs)
        self.store_password = store_password

    def add_prefix(self, field_name):
        # Look up field name; return original if not found
        field_name = self.FIELD_NAME_LOOKUP.get(
            field_name, field_name
        )
        return super(SunTechForm, self).add_prefix(field_name)


class SunTechSendForm(SunTechForm):
    """
    A generic form subclassing `SunTechForm` but transform all fields in it
    to be hidden. This class is abstract in concept and should be subclassed.
    """
    store_name = forms.CharField()
    price = forms.IntegerField()
    order_id = forms.IntegerField()
    order_info = forms.CharField(required=False)
    name = forms.CharField()
    phone = NumberStringField()
    email = forms.EmailField(required=False)
    note1 = forms.CharField(required=False)
    note2 = forms.CharField(required=False)
    checksum = SunTechChecksumField()

    FIELD_NAME_LOOKUP = {
        'store_name': 'web',
        'price': 'MN',
        'order_id': 'Td',
        'order_info': 'OrderInfo',
        'name': 'sna',
        'phone': 'sdt',
        'email': 'email',
        'note1': 'note1',
        'note2': 'note2',
        'checksum': 'ChkValue'
    }

    def __init__(self, *args, **kwargs):
        super(SunTechSendForm, self).__init__(*args, **kwargs)
        for key in self.fields:
            field = self.fields[key]
            if isinstance(field, forms.IntegerField):
                field.widget = HiddenIntegerInput()
            else:
                field.widget = widgets.HiddenInput()

    def get_checksum(self):
        """
        Subclasses may override this method to provide another valid algorithm
        to generate and return a checksum from `self['field_name'].value()` and
        `self.store_password`. This is called by `fill_checksum`.
        """
        return _sha1_hex_upper(
            self['store_name'].value(),
            self.store_password,
            int(self['price'].value())
        )

    def fill_checksum(self, field_name='checksum'):
        """
        Fills the checksum field with get_checksum. The checksum field is named
        "checksum" by default, but you can pass in another name if needed.
        """
        self.fields[field_name].initial = self.get_checksum()


class SunTechReceiveForm(SunTechForm):
    """
    A generic form subclassing `SunTechForm` to provide `validate_checksum()`
    method. Subclasses should override this method to check for the form's
    checksum based on values in `self.cleaned_data` and `self.store_password`,
    raising a `ChecksumError` if the checksum is not correct.

    `validate_checksum` is called in `clean`. Therefore, all subclasses of
    `SunTechReceiveForm` should invoke `super` in their `clean` overrides.
    """
    sun_id = forms.CharField()
    store_name = forms.CharField()
    order_id = forms.IntegerField()
    price = forms.IntegerField()
    note1 = forms.CharField(required=False)
    note2 = forms.CharField(required=False)
    checksum = SunTechChecksumField()

    FIELD_NAME_LOOKUP = {
        'sun_id': 'buysafeno',
        'store_name': 'web',
        'order_id': 'Td',
        'price': 'MN',
        'note1': 'note1',
        'note2': 'note2',
        'checksum': 'ChkValue'
    }

    class SEND_TYPE:
        BACKGROUND = 1
        WEBPAGE = 2

    def validate_checksum(self, field_name='checksum'):
        reference = _sha1_hex_upper(
            self.cleaned_data['store_name'], self.store_password,
            self.cleaned_data['sun_id'], self.cleaned_data['price'],
            self.cleaned_data['return_code']
        )
        checksum = self.cleaned_data['checksum']
        if reference != checksum:
            raise ChecksumError(expected=reference, got=checksum)

    def clean(self):
        self.validate_checksum()
        return super(SunTechReceiveForm, self).clean()


SUNTECH_INSTANT_RESPONSE_LOOKUP = deepcopy(
    SunTechReceiveForm.FIELD_NAME_LOOKUP
)
SUNTECH_INSTANT_RESPONSE_LOOKUP.update({
    'name': 'Name',
    'site_name': 'webname',
    'send_type': 'SendType',
    'return_code': 'errcode',
    'return_detail': 'errmsg',
})


class SunTechInstantResponseForm(SunTechReceiveForm):
    """
    A generic form subclassing `SunTechReceiveForm` to be base of receive forms
    for payment methods with an "Instant Response," i.e. gets a response almost
    instantly because customer pays directly on the Internet.
    """
    name = forms.CharField()
    site_name = forms.CharField()
    send_type = forms.IntegerField(
        min_value=SunTechReceiveForm.SEND_TYPE.BACKGROUND,
        max_value=SunTechReceiveForm.SEND_TYPE.WEBPAGE
    )
    return_code = forms.CharField()
    return_detail = forms.CharField(required=False)

    FIELD_NAME_LOOKUP = SUNTECH_INSTANT_RESPONSE_LOOKUP

    def clean_name(self):
        name = urlunquote(self.cleaned_data['name'])
        return name

    def clean_site_name(self):
        site_name = urlunquote(self.cleaned_data['site_name'])
        return site_name

    def clean_return_detail(self):
        detail = urlunquote(self.cleaned_data['return_detail'])
        return detail

    def is_successful(self):
        return (self.cleaned_data.get('return_code') == '00')


SUNTECH_PAY_LATER_SEND_LOOKUP = deepcopy(SunTechSendForm.FIELD_NAME_LOOKUP)
SUNTECH_PAY_LATER_SEND_LOOKUP.update({
    'due_date': 'DueDate',
    'bill_date': 'BillDate',
    'user_id': 'UserNo'
})


class SunTechPayLaterSendForm(SunTechSendForm):
    """
    A generic form subclassing `SunTechSendForm` to be base of send forms for
    payment methods without an instant response. This kind of pyments usually
    consists a manual payment collection, such as offline ATM transactions.
    """
    due_date = forms.DateField(input_formats=["%Y%m%d"])
    bill_date = forms.DateField(input_formats=["%Y%m%d"])
    user_id = forms.IntegerField()

    FIELD_NAME_LOOKUP = SUNTECH_PAY_LATER_SEND_LOOKUP


SUNTECH_PAY_LATER_RECEIVE_LOOKUP = deepcopy(
    SunTechReceiveForm.FIELD_NAME_LOOKUP
)
SUNTECH_PAY_LATER_RECEIVE_LOOKUP.update({
    'return_code': 'errcode',
    'user_id': 'UserNo',
    'payment_date': 'PayDate',
    'payment_type': 'PayType'
})


class SunTechPayLaterReceiveForm(SunTechReceiveForm):
    """
    A generic form subclassing `SunTechReceiveForm` to be base of receive forms
    for payment methods without an instant response. This kind of pyments
    usually consists a manual payment collection, such as offline ATM
    transactions.
    """
    class PAYMENT_TYPE:
        BARCODE_CONVINIENCE_STORE = 1
        BARCODE_POST_OFFICE = 2
        VIRTUAL_ACCOUNT = 3
        CASH_CONVINIENCE_STORE = 4

    return_code = forms.CharField()
    user_id = forms.IntegerField()
    payment_date = forms.DateField(input_formats=["%Y%m%d"])
    payment_type = forms.IntegerField(
        min_value=PAYMENT_TYPE.BARCODE_CONVINIENCE_STORE,
        max_value=PAYMENT_TYPE.CASH_CONVINIENCE_STORE
    )

    FIELD_NAME_LOOKUP = SUNTECH_PAY_LATER_RECEIVE_LOOKUP


#################
# BuySafe forms #
#################

class BuySafeSendForm(SunTechSendForm):
    pass


BUYSAFE_RECEIVE_LOOKUP = deepcopy(SunTechInstantResponseForm.FIELD_NAME_LOOKUP)
BUYSAFE_RECEIVE_LOOKUP.update({
    'auth_code': 'ApproveCode',
    'card_number_tail': 'Card_NO',
})


class BuySafeReceiveForm(SunTechInstantResponseForm):
    auth_code = NumberStringField(required=False)
    card_number_tail = NumberStringField(max_length=4, required=False)

    FIELD_NAME_LOOKUP = BUYSAFE_RECEIVE_LOOKUP

    def clean(self):
        self.cleaned_data = super(BuySafeReceiveForm, self).clean()

        # Validate fields that are only available when return_code is 00
        if self.cleaned_data['return_code'] == '00':
            if not self.cleaned_data.get('auth_code', None):
                raise forms.ValidationError(
                    'Stated as successful, but auth code not presented.'
                )
            if len(self.cleaned_data.get('card_number_tail', '')) != 4:
                raise forms.ValidationError(
                    'Stated as successful, but credit card number not '
                    'presented.'
                )
        return self.cleaned_data


#################
# Web ATM forms #
#################

class WebATMSendForm(SunTechSendForm):
    pass


class WebATMReceiveForm(SunTechInstantResponseForm):
    pass


###############
# 24Pay forms #
###############

# class Pay24FormMeta(forms.DeclarativeFieldsMetaclass):
#     """
#     Metaclass to generate product name, price and quantity fields
#     """
#     def __new__(cls, name, bases, attrs):
#         for suffix, field_type in zip(
#                 ['name', 'price', 'quantity'],
#                 [forms.CharField, forms.DecimalField, forms.IntegerField]
#         ):
#             for i in range(10):
#                 name = '_'.join(['product', suffix, str(i)])
#                 attrs[name] = field_type(required=False)
#         return forms.DeclarativeFieldsMetaclass.__new__(
#             cls, name, bases, attrs
#         )


# PAY24_SEND_LOOKUP = deepcopy(SunTechPayLaterSendForm.FIELD_NAME_LOOKUP)
# for f, h in zip(['name', 'unit_price', 'quantity'],
#                 ['Name', 'Price', 'Quantity']):
#     for i in range(10):
#         fn = '_'.join(['product', f, str(i)])
#         hn = ''.join(['Product', h, str(i + 1)])
#         PAY24_SEND_LOOKUP[fn] = hn


# class Pay24SendForm(SunTechPayLaterSendForm):
#     product_count = 0

#     FIELD_NAME_LOOKUP = PAY24_SEND_LOOKUP

#     def add_product(self, product_name, )


# class Pay24ReceiveForm(SunTechPayLaterReceiveForm):
#     pass


#################
# PayCode forms #
#################

class PayCodeSendForm(SunTechPayLaterSendForm):
    pass


PAY_CODE_INSTANT_LOOKUP = deepcopy(SunTechReceiveForm.FIELD_NAME_LOOKUP)
PAY_CODE_INSTANT_LOOKUP.update({
    'user_id': 'UserNo',
    'payment_id': 'paycode'
})


class PayCodeInstantReceiveForm(SunTechReceiveForm):
    user_id = forms.IntegerField()
    payment_id = forms.CharField()

    FIELD_NAME_LOOKUP = PAY_CODE_INSTANT_LOOKUP


class PayCodeReceiveForm(SunTechPayLaterReceiveForm):
    pass



from django.forms import widgets


class NumberTextInput(widgets.TextInput):
    """
    A TextInput subclass that omits all non-digit characters when rendering.
    """
    def render(self, name, value, attrs=None):
        trimmed = ''
        for char in unicode(value):
            if char.isdigit():
                trimmed += char
        return super(NumberTextInput, self).render(name, trimmed, attrs)


class HiddenIntegerInput(widgets.HiddenInput):
    """
    A HiddenInput that ensures that the value is an integer.
    """
    def render(self, name, value, attrs=None):
        return super(HiddenIntegerInput, self).render(name, int(value), attrs)



# -*- coding: utf8

from django.http import HttpResponseBadRequest, HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
from django.shortcuts import render
from buysafe.models import PaymentMethod
from buysafe.forms import SunTechReceiveForm, BuySafeSendForm, WebATMSendForm
from buysafe.utils import (
    call_handler, call_handler_and_render, get_payment_form,
    default_order_info_handler, make_response_handler
)


PAYMENT_SEND_FORMS = {
    PaymentMethod.TYPE_BUYSAFE: BuySafeSendForm,
    PaymentMethod.TYPE_WEBATM: WebATMSendForm
}


def entry(request, order_id, template='buysafe/entry.html'):
    return render(request, template, {'order_id': order_id})


@require_POST
def start(request, template='buysafe/start.html'):
    keyword_args = {}
    for k in request.POST:
        keyword_args[k] = request.POST[k]

    context = {}

    info = call_handler(
        'BUYSAFE_FORM_VALUES_GENERATOR',
        default_order_info_handler,
        request=request, context=context, **keyword_args
    )

    forms = []
    payment_methods = PaymentMethod.enabled.all()
    for method in payment_methods:
        form_model = PAYMENT_SEND_FORMS[method.payment_type]
        info['store_name'] = method.store_id
        form = form_model(
            store_password=method.password, initial=info
        )
        form.fill_checksum()
        form.submit_button_title = method.name
        forms.append(form)

    context['forms'] = forms
    return call_handler_and_render(
        'BUYSAFE_START_HANDLER', None,
        request=request, template=template, context=context,
        forms=forms, payment_methods=payment_methods
    )


@csrf_exempt
@require_POST
def success(request, payment_type, template='buysafe/success.html'):
    context = {}
    payment_type = int(payment_type)
    form = get_payment_form(payment_type, request.POST)
    if form is None:
        return call_handler_and_render(
            'BUYSAFE_SUCCESS_INVALID_HANDLER',
            make_response_handler(HttpResponseBadRequest),
            request=request, context=context, form=form
        )
    context['data'] = form.cleaned_data
    send_type = form.cleaned_data['send_type']
    if send_type == SunTechReceiveForm.SEND_TYPE.BACKGROUND:
        return call_handler_and_render(
            'BUYSAFE_SUCCESS_BACKGROUND_HANDLER',
            make_response_handler(HttpResponse),
            request=request, context=context, form=form
        )
    return call_handler_and_render(
        'BUYSAFE_SUCCESS_RENDER_HANDLER', None,
        request=request, template=template, context=context, form=form
    )


@csrf_exempt
@require_POST
def fail(request, payment_type, template='buysafe/fail.html'):
    context = {}
    payment_type = int(payment_type)
    form = get_payment_form(payment_type, request.POST)
    if form is None:
        return call_handler_and_render(
            'BUYSAFE_FAIL_INVALID_HANDLER',
            make_response_handler(HttpResponseBadRequest),
            request=request, context=context, form=form
        )
    context['data'] = form.cleaned_data
    send_type = form.cleaned_data['send_type']
    if send_type == SunTechReceiveForm.SEND_TYPE.BACKGROUND:
        return call_handler_and_render(
            'BUYSAFE_FAIL_BACKGROUND_HANDLER',
            make_response_handler(HttpResponse),
            request=request, context=context, form=form
        )
    return call_handler_and_render(
        'BUYSAFE_FAIL_RENDER_HANDLER', None,
        request=request, template=template, context=context, form=form
    )


@csrf_exempt
@require_POST
def check(request, payment_type):
    context = {}
    payment_type = int(payment_type)
    form = get_payment_form(payment_type, request.POST)
    if form is None:
        return call_handler_and_render(
            'BUYSAFE_CHECK_INVALID_HANDLER',
            make_response_handler(HttpResponseBadRequest),
            request=request, context=context, form=form
        )
    send_type = form.cleaned_data['send_type']
    if send_type == SunTechReceiveForm.SEND_TYPE.BACKGROUND:
        return call_handler_and_render(
            'BUYSAFE_CHECK_HANDLER',
            make_response_handler(HttpResponse, '0000'),
            request=request, context=context, form=form
        )
        return HttpResponse('0000')
    return call_handler_and_render(
        'BUYSAFE_CHECK_INVALID_HANDLER',
        make_response_handler(HttpResponseBadRequest),
        request=request, context=context, form=form
    )



import re
from django import forms
from buysafe.widgets import NumberTextInput


class NumberStringField(forms.CharField):
    """
    A CharField subclass that contains only numbers. This is different with
    IntegerField because NumberStringField can have leading zeros. Usaful for
    Things like telephone numbers
    """
    widget = NumberTextInput

    def validate(self, value):
        super(NumberStringField, self).validate(value)
        if re.match(r'^\d*$', value) is None:
            raise forms.ValidationError(
                'This field should only consist numbers'
            )


class SunTechChecksumField(forms.CharField):
    def validate(self, value):
        super(SunTechChecksumField, self).validate(value)
        if re.match(r'^[0-9A-F]{40}$', value) is None:
            raise forms.ValidationError(
                'Checksum should be of length 40, and consists only numbers '
                'and/or capital alphabets from A through F.'
            )
